@@ -5,14 +5,13 @@
 #include "HttpRequest.h"
 #include "HttpReply.h"
 #include "BodyPipe.h"
+#include "TextException.h"
 #include <libecap/common/names.h>
 #include <libecap/common/area.h>
 #include <libecap/common/version.h>
-#include <libecap/common/named_values.h>
 #include "adaptation/ecap/MessageRep.h"
 #include "adaptation/ecap/XactionRep.h"
 #include "adaptation/ecap/Host.h" /* for protocol constants */
-#include "base/TextException.h"
 
 /* HeaderRep */
 
@@ -49,9 +48,6 @@ Adaptation::Ecap::HeaderRep::add(const Name &name, const Value &value)
     HttpHeaderEntry *e = new HttpHeaderEntry(squidId, name.image().c_str(),
             value.toString().c_str());
     theHeader.addEntry(e);
-
-    if (squidId == HDR_CONTENT_LENGTH)
-        theMessage.content_length = theHeader.getInt64(HDR_CONTENT_LENGTH);
 }
 
 void
@@ -62,20 +58,6 @@ Adaptation::Ecap::HeaderRep::removeAny(const Name &name)
         theHeader.delByName(name.image().c_str());
     else
         theHeader.delById(squidId);
-
-    if (squidId == HDR_CONTENT_LENGTH)
-        theMessage.content_length = theHeader.getInt64(HDR_CONTENT_LENGTH);
-}
-
-void
-Adaptation::Ecap::HeaderRep::visitEach(libecap::NamedValueVisitor &visitor) const
-{
-    HttpHeaderPos pos = HttpHeaderInitPos;
-    while (HttpHeaderEntry *e = theHeader.getEntry(&pos)) {
-        const Name name(e->name.termedBuf()); // optimize: find std Names
-        name.assignHostId(e->id);
-        visitor.visit(name, Value(e->value.rawBuf(), e->value.size()));
-    }
 }
 
 libecap::Area
@@ -136,40 +118,38 @@ Adaptation::Ecap::FirstLineRep::protocol() const
 {
     // TODO: optimize?
     switch (theMessage.protocol) {
-    case AnyP::PROTO_HTTP:
+    case PROTO_HTTP:
         return libecap::protocolHttp;
-    case AnyP::PROTO_HTTPS:
+    case PROTO_HTTPS:
         return libecap::protocolHttps;
-    case AnyP::PROTO_FTP:
+    case PROTO_FTP:
         return libecap::protocolFtp;
-    case AnyP::PROTO_GOPHER:
+    case PROTO_GOPHER:
         return libecap::protocolGopher;
-    case AnyP::PROTO_WAIS:
+    case PROTO_WAIS:
         return libecap::protocolWais;
-    case AnyP::PROTO_WHOIS:
+    case PROTO_WHOIS:
         return libecap::protocolWhois;
-    case AnyP::PROTO_URN:
+    case PROTO_URN:
         return libecap::protocolUrn;
-    case AnyP::PROTO_ICP:
+    case PROTO_ICP:
         return protocolIcp;
 #if USE_HTCP
-    case AnyP::PROTO_HTCP:
+    case PROTO_HTCP:
         return protocolHtcp;
 #endif
-    case AnyP::PROTO_CACHE_OBJECT:
+    case PROTO_CACHEOBJ:
         return protocolCacheObj;
-    case AnyP::PROTO_INTERNAL:
+    case PROTO_INTERNAL:
         return protocolInternal;
-    case AnyP::PROTO_ICY:
-        return protocolIcy;
-    case AnyP::PROTO_UNKNOWN:
-        return protocolUnknown; // until we remember the protocol image
-    case AnyP::PROTO_NONE:
+    case PROTO_ICY:
+        return Name();
+    case PROTO_NONE:
         return Name();
 
-    case AnyP::PROTO_MAX:
+    case PROTO_MAX:
         break; // should not happen
-        // no default to catch AnyP::PROTO_ additions
+        // no default to catch PROTO_ additions
     }
     Must(false); // not reached
     return Name();
@@ -182,12 +162,12 @@ Adaptation::Ecap::FirstLineRep::protocol(const Name &p)
     theMessage.protocol = TranslateProtocolId(p);
 }
 
-AnyP::ProtocolType
+protocol_t
 Adaptation::Ecap::FirstLineRep::TranslateProtocolId(const Name &name)
 {
     if (name.assignedHostId())
-        return static_cast<AnyP::ProtocolType>(name.hostId());
-    return AnyP::PROTO_UNKNOWN;
+        return static_cast<protocol_t>(name.hostId());
+    return PROTO_NONE; // no PROTO_OTHER
 }
 
 
@@ -213,10 +193,8 @@ Adaptation::Ecap::RequestLineRep::uri(const Area &aUri)
 Adaptation::Ecap::RequestLineRep::Area
 Adaptation::Ecap::RequestLineRep::uri() const
 {
-    const char *fullUrl = urlCanonical(&theMessage);
-    Must(fullUrl);
-    // optimize: avoid copying by having an Area::Detail that locks theMessage
-    return Area::FromTempBuffer(fullUrl, strlen(fullUrl));
+    return Area::FromTempBuffer(theMessage.urlpath.rawBuf(),
+                                theMessage.urlpath.size());
 }
 
 void
